%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(sl_alloc_SUITE).
-export([all/1]).
-export([mmap_table/1,
	 param/1,
	 grow_heap/1,
	 grow_stack/1,
	 grow_stack_heap/1,
	 bucket_index/1,
	 coalesce/1,
	 coalesce_debug/1,
	 sl_realloc_copy/1,
	 old_sl_realloc_copy/1,
	 sl_realloc_copy_debug/1,
	 old_sl_realloc_copy_debug/1,
	 threads/1,
	 old_threads/1,
	 threads_debug/1,
	 old_threads_debug/1,
	 old_mmap_leak/1,
	 old_mmap_leak_debug/1,
	 old_mmap_max/1,
	 old_mmap_max_debug/1,
	 halloc_wrap/1,
	 skipped_cases/1]).

-define(SL_ALLOC_REL, "2").
-define(OLD_SL_ALLOC_REL, "1").

-define(PP_SUCCEEDED, 0).
-define(PP_FAILED,  1).
-define(PP_SKIPPED, 17).

-define(ALL_CASES_ENV, "RUN_ALL_SL_ALLOC_TEST_CASES").

-include("test_server.hrl").

all(doc) -> [];
all(suite) ->
    cases() ++ case os:getenv(?ALL_CASES_ENV) of
		   false -> [skipped_cases];
		   _     -> extra_cases()
	       end.


extra_cases() ->
    [mmap_table,
     grow_heap,
     grow_stack,
     grow_stack_heap,
     old_mmap_leak,
     old_mmap_leak_debug,
     old_mmap_max,
     old_mmap_max_debug,
     old_sl_realloc_copy,
     old_sl_realloc_copy_debug,
     old_threads,
     old_threads_debug].    

cases() ->
    [param,
     bucket_index,
     coalesce,
     coalesce_debug,
     sl_realloc_copy,
     sl_realloc_copy_debug,
     threads,
     threads_debug,
     halloc_wrap].

skipped_cases(doc) -> [];
skipped_cases(suite) -> [];
skipped_cases(Config) ->
    {skipped,
     lists:flatten(io_lib:format("Skipped test cases: ~w. "
				 "Set environment variable "
				 ?ALL_CASES_ENV
				 " in order to run these test "
				 "cases.",
				 [extra_cases()]))}.

param(doc) ->
    ["Test that memory allocation command line options are interpreted"
     " correct by sl_alloc "];
param(suite) ->
    [];
param(Config) when list(Config) ->
    case os:type() of 
	vxworks ->
	    {skipped, "Not run on vxworks"};
	_ ->
	    run_param(Config)
    end.

run_param(Config) ->
    Dog = test_server:timetrap(test_server:seconds(60)),
    ?line test_server:format("Test server sl_alloc info:~n~p~n",
			     [erlang:system_info(sl_alloc)]),
    ?line {A,V,F,S} = erlang:system_info(allocator),
    ?line test_server:format("Test server allocator ~p~n", [{A,V,F,S}]),
    ?line [Name1, Name2, Name3] = get_nodenames(3, param),
    ?line Self = self(),
    ?line {ok, Node1} = start_node(Name1,
				   "+Se true +Sr " ?SL_ALLOC_REL
				   " +Ssbct 512 +Ssbcmt 50 +Smmc 10 "
				   "+Smcs 5120 +Sscs 2000 +Slcs 30000 "
				   "+Smbcgs 50 +Smsbclt 60 +Smbsd 10"),
    ?line Ref1 = make_ref(),
    ?line spawn_link(Node1,
		     fun () ->
			     Self ! {erlang:system_info(allocator), Ref1}
		     end),
    receive
	{{_, _, FeatList1, SettList1} = Alloc1, Ref1} ->
	    ?line stop_node(Node1),
	    ?line test_server:format("~p sl_alloc info:~n~p~n",
				     [Node1,Alloc1]),
	    ?line SlaS1 = fetch(sl_alloc, SettList1),
	    ?line true = lists:member(sl_alloc, FeatList1),
	    ?line true = fetch(enabled, SlaS1),
	    ?line 512*1024 = fetch(singleblock_carrier_threshold, SlaS1),
	    ?line 50 = fetch(singleblock_carrier_move_threshold, SlaS1),
	    ?line 60 = fetch(mmap_singleblock_carrier_load_threshold, SlaS1),
	    ?line case lists:member(mmap, FeatList1) of
		      true ->
			  ?line 10 = fetch(max_mmap_carriers, SlaS1);
		      false ->
			  ?line 0 = fetch(max_mmap_carriers, SlaS1)
		  end,
	    ?line 5120*1024 = fetch(main_carrier_size, SlaS1),
	    ?line 2000*1024 = fetch(smallest_multiblock_carrier_size, SlaS1),
	    ?line 30000*1024 = fetch(largest_multiblock_carrier_size, SlaS1),
	    ?line 50 = fetch(multiblock_carrier_growth_stages, SlaS1),
	    ?line 10 = fetch(max_block_search_depth, SlaS1)
    end,
    ?line {ok, Node2} = start_node(Name2,
				   "+Se true +Ssbct 512 +Smmc 10 +Sr "
				   ?OLD_SL_ALLOC_REL),
    ?line Ref2 = make_ref(),
    ?line spawn_link(Node2,
		     fun () ->
			     Self ! {erlang:system_info(allocator), Ref2},
			     unlink(Self)
		     end),
    receive
	{{_, _, FeatList2, SettList2} = Alloc2, Ref2} ->
	    ?line stop_node(Node2),
	    ?line test_server:format("~p allocator:~n~p~n",
				     [Node2,Alloc2]),
	    ?line SlaS2 = fetch(sl_alloc, SettList2),
	    ?line true = lists:member(sl_alloc, FeatList2),
	    ?line case lists:member(mmap, FeatList2) of
		      true ->
			  ?line true = fetch(enabled, SlaS2),
			  ?line 512*1024 =
			      fetch(singleblock_carrier_threshold, SlaS2),
			  ?line 10 = fetch(max_mmap_carriers, SlaS2);
		      false ->
			  ?line ok
		  end
    end,
    ?line {ok, Node3} = start_node(Name3, "+Se false"),
    ?line Ref3 = make_ref(),
    ?line spawn_link(Node3,
		     fun () ->
			     Self ! {erlang:system_info(allocator), Ref3},
			     unlink(Self)
		     end),
    receive
	{{_, _, FeatList3, SettList3} = Alloc3, Ref3} ->
	    ?line stop_node(Node3),
	    ?line test_server:format("~p allocator:~n~p~n",
				     [Node3,Alloc3]),
	    ?line false = lists:member(sl_alloc, FeatList3),
	    ?line false = fetch(enabled, fetch(sl_alloc, SettList3))
    end,
    ok.

mmap_table(suite) -> [];
mmap_table(doc) -> ["Force usage of sl_alloc's mmap_table (disable regions"
		    "optimization) and provoke a lot of mmaps."];
mmap_table(Config) when list(Config) -> {skipped, "Not implemented yet"}.


% The grow_heap, grow_stack, and grow_stack_heap test cases are stolen
% from the gc_SUITE and then modified.

grow_heap(suite) -> [];
grow_heap(doc) -> ["Produce a growing list of elements, ",
		   "for X calls, then drop one item per call",
		   "until the list is empty."];
grow_heap(Config) when list(Config) ->
    run_grow_heap("+Smmc 10000 +Smcs 40 +Sscs 40 +Ssbct 4 "
		  "+Se true +Sr " ?SL_ALLOC_REL),
    run_grow_heap("+Se true +Sr " ?SL_ALLOC_REL),
    run_grow_heap("+Smmc 10000 +Ssbct 4 +Se true +Sr " ?OLD_SL_ALLOC_REL),
    ok.

run_grow_heap(NodeArgs) ->
    ?line Dog=test_server:timetrap(test_server:minutes(40)),
    ?line [Name] = get_nodenames(1, grow_heap),
    ?line ?t:format("Running test with args: ~s", [NodeArgs]),
    ?line {ok, Node} = start_node(Name, NodeArgs),
    ?line Self = self(),
    ?line Ref = make_ref(),
    Worker = spawn_link(Node,
			fun () ->
				?line print_sl_alloc_info(),
				?line ok=grow_heap1(100),
				?line print_sl_alloc_info(),
				?line ok=grow_heap1(5000),
				?line print_sl_alloc_info(),
				?line ok=grow_heap1(100),
				?line print_sl_alloc_info(),
				?line ok=grow_heap1(15000),
				?line print_sl_alloc_info(),
				Self ! Ref,
				unlink(Self)
			end),
    receive
	Ref ->
	    ?line unlink(Worker),
	    ?line stop_node(Node)
    end,
    ?line test_server:timetrap_cancel(Dog),
    ok.

grow_heap1(Len) ->
    io:format("~ngrow_heap with ~p items.",[Len]),
    show_heap("before:"),
    grow_heap1([], Len, 0, up),
    show_heap("after:").

grow_heap1(List, MaxLen, MaxLen, up) ->
    show_heap("top:"),
    grow_heap1(List, MaxLen, MaxLen-1, down);
grow_heap1(List, MaxLen, CurLen, up) ->
    NewList=[make_arbit()|List],
    grow_heap1(NewList, MaxLen, CurLen+1, up);
grow_heap1([], MaxLen, _, down) ->
    ok;
grow_heap1([Elem0|List], MaxLen, CurLen, down) ->
    ?line {A,B,C}=erlang:now(),
    ?line Num=C rem (length(List))+1,
    ?line Elem=lists:nth(Num, List),
    ?line NewList=lists:delete(Elem, List),
    grow_heap1(NewList, MaxLen, CurLen-1, down).


grow_stack(suite) -> [];
grow_stack(doc) -> ["Increase and decrease stack size, and ",
		    "drop off some garbage from time to time."];
grow_stack(Config) when list(Config) ->
    run_grow_stack("+Smmc 10000 +Smcs 40 +Sscs 40 +Ssbct 4 "
		   "+Se true +Sr " ?SL_ALLOC_REL),
    run_grow_stack("+Se true +Sr " ?SL_ALLOC_REL),
    run_grow_stack("+Smmc 10000 +Ssbct 4 +Se true +Sr " ?OLD_SL_ALLOC_REL),
    ok.

run_grow_stack(NodeArgs) ->
    ?line Dog=test_server:timetrap(test_server:minutes(80)),
    ?line [Name] = get_nodenames(1, grow_stack),
    ?line ?t:format("Running test with args: ~s", [NodeArgs]),
    ?line {ok, Node} = start_node(Name, NodeArgs),
    ?line Self = self(),
    ?line Ref = make_ref(),
    Worker = spawn_link(Node,
			fun () ->
				?line print_sl_alloc_info(),
				show_heap("before:"),
				?line grow_stack1(10, 0),
				show_heap("after:"),
				?line print_sl_alloc_info(),
				show_heap("before:"),
				?line grow_stack1(100, 0),
				show_heap("after:"),
				?line print_sl_alloc_info(),
				show_heap("before:"),
				?line grow_stack1(300, 0),
				show_heap("after:"),
				?line print_sl_alloc_info(),
				show_heap("before:"),
				?line grow_stack1(500, 0),
				show_heap("after:"),
				?line print_sl_alloc_info(),
				Self ! Ref,
				unlink(Self)
			end),
    receive
	Ref ->
	    ?line unlink(Worker),
	    ?line stop_node(Node)
    end,
    ?line test_server:timetrap_cancel(Dog),
    ok.

grow_stack1(0, _) ->
    ok;
grow_stack1(Recs, 0) ->
%    show_heap("running:"),
    grow_stack1(Recs-1, Recs),
    grow_stack1(0,0);
grow_stack1(Recs, CurRecs) ->
    grow_stack1(Recs, CurRecs-1),
    make_arbit(),
    grow_stack1(1,0),
    ok.


%% Let's see how BEAM handles this one...
grow_stack_heap(suite) -> [];
grow_stack_heap(doc) -> ["While growing the heap, bounces the size ",
			 "of the stack, and while reducing the heap",
			 "bounces the stack usage."];
grow_stack_heap(Config) when list(Config) ->
    run_grow_stack_heap("+Smmc 10000 +Smcs 40 +Sscs 40 +Ssbct 4 "
			"+Se true +Sr "	?SL_ALLOC_REL),
    run_grow_stack_heap("+Se true +Sr " ?SL_ALLOC_REL),
    run_grow_stack_heap("+Smmc 10000 +Ssbct 4 +Se true +Sr " ?OLD_SL_ALLOC_REL),
    ok.

run_grow_stack_heap(NodeArgs) ->
    ?line Dog=test_server:timetrap(test_server:minutes(40)),
    ?line [Name] = get_nodenames(1, grow_stack),
    ?line ?t:format("Running test with args: ~s", [NodeArgs]),
    ?line {ok, Node} = start_node(Name, NodeArgs),
    ?line Self = self(),
    ?line Ref = make_ref(),
    Worker = spawn_link(Node,
			fun () ->
				?line print_sl_alloc_info(),
				?line grow_stack_heap1(24),
				?line print_sl_alloc_info(),
				?line grow_stack_heap1(50),
				?line print_sl_alloc_info(),
				?line grow_stack_heap1(76),
				?line print_sl_alloc_info(),
				Self ! Ref,
				unlink(Self)
			end),
    receive
	Ref ->
	    ?line unlink(Worker),
	    ?line stop_node(Node)
    end,
    ?line test_server:timetrap_cancel(Dog),
    ok.

grow_stack_heap1(MaxLen) ->
    io:format("~ngrow_stack_heap with ~p items.",[MaxLen]),
    show_heap("before:"),
    grow_stack_heap1([], MaxLen, 0, up),
    show_heap("after:").

grow_stack_heap1(List, MaxLen, MaxLen, up) ->
    show_heap("top:"),
    grow_stack_heap1(List, MaxLen, MaxLen-1, down);
grow_stack_heap1(List, MaxLen, CurLen, up) ->
    grow_stack1(CurLen*2,0),
    grow_stack_heap1([make_arbit()|List], MaxLen, CurLen+1, up),
    ok;

grow_stack_heap1([], MaxLen, _, down) ->
    ok;
grow_stack_heap1([Elem0|List], MaxLen, CurLen, down) ->
    grow_stack1(CurLen*2,0),
    ?line {A,B,C}=erlang:now(),
    ?line Num=C rem (length(List))+1,
    ?line Elem=lists:nth(Num, List),
    ?line NewList=lists:delete(Elem, List),
    grow_stack_heap1(NewList, MaxLen, CurLen-1, down),
    ok.


%% Create an arbitrary element/term.
make_arbit() ->
    {AA,BB,CC}=erlang:now(),
    A=AA+1, B=BB+1, C=CC+1,
    New =
	case C rem 9 of
	    0 -> make_string((B div C) +5);
	    1 -> C;
	    2 -> make_ref();
	    3 -> self();
	    4 -> list_to_binary(make_string((C div B) + 12));
	    5 -> (C*B)/(A+1);
	    6 -> list_to_tuple(make_string((B div C) +5));
	    7 -> list_to_atom(make_string(((C div B) rem 254) + 2));
	    8 -> fun(X) -> {X,AA,make_string((B div C)+10)} end
	end,
    New.

%% Create an arbitrary string of a certain length.
make_string(Length) ->
    Alph="abcdefghjiklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"++
	"0123456789",
    make_string(Alph, Length, []).

make_string(_, 0, Acc) ->
    Acc;
make_string(Alph, Length, Acc) ->
    {A,B,C}=erlang:now(),
    Pos=1+(Length*C rem length(Alph)),
    make_string(Alph, Length-1, 
		[lists:nth(Pos,Alph)|Acc]).

show_heap(String) ->
    garbage_collect(self()),
    receive after 1 -> ok end,
    {heap_size, HSize}=process_info(self(), heap_size),
    {stack_size, SSize}=process_info(self(), stack_size),
    io:format("Heap/Stack "++String++"~p/~p", [HSize, SSize]).

bucket_index(suite) -> [];
bucket_index(doc) ->   [];
bucket_index(Config) when list(Config) ->
    run_pp_test_case("bucket_index", Config, opt, 5*60).

coalesce(suite) -> [];
coalesce(doc) ->   ["Checks that sl_realloced data is copied as"
			   "it should"];
coalesce(Config) when list(Config) ->
    run_pp_test_case("coalesce", Config, opt, 5*60).

coalesce_debug(suite) -> [];
coalesce_debug(doc) ->   ["Checks that sl_realloced data is copied as"
			   "it should"];
coalesce_debug(Config) when list(Config) ->
    run_pp_test_case("coalesce", Config, debug, 5*60).

sl_realloc_copy(suite) -> [];
sl_realloc_copy(doc) ->   ["Checks that sl_realloced data is copied as"
			   "it should"];
sl_realloc_copy(Config) when list(Config) ->
    run_pp_test_case("sl_realloc_copy", Config, opt, 5*60).

old_sl_realloc_copy(suite) -> [];
old_sl_realloc_copy(doc) ->   ["Checks that sl_realloced data is copied as"
			       "it should"];
old_sl_realloc_copy(Config) when list(Config) ->
    run_pp_test_case("sl_realloc_copy", "old_sl_alloc", Config, opt, 5*60).

sl_realloc_copy_debug(suite) -> [];
sl_realloc_copy_debug(doc) ->   ["Checks that sl_realloced data is copied as"
				 "it should (debug compiled)"];
sl_realloc_copy_debug(Config) when list(Config) ->
    run_pp_test_case("sl_realloc_copy", Config, debug, 10*60).

old_sl_realloc_copy_debug(suite) -> [];
old_sl_realloc_copy_debug(doc) ->   ["Checks that sl_realloced data is "
				     "copied as it should (debug compiled)"];
old_sl_realloc_copy_debug(Config) when list(Config) ->
    run_pp_test_case("sl_realloc_copy","old_sl_alloc",Config,debug,10*60).

threads_tests_modifier() ->
    case os:type() of
	{unix, Flavor} when Flavor == freebsd;
			    Flavor == darwin ->
	    {{comment,
	      "No fork safety test was performed since the pthreads "
	      "implementation on " ++ atom_to_list(Flavor) ++ " lacks "
	      "the pthread_atfork() function and therefore is known not"
	      " to be fork safe."},
	     "no_fork_test"};
	_ ->
	    {ok, ""}
    end.

threads(suite) -> [];
threads(doc) ->   ["Checks that sl_alloc function okay in a threaded"
		   "environment"];
threads(Config) when list(Config) ->
    {OkRes, PPArg} = threads_tests_modifier(),
    case run_pp_test_case("threads", "new_sl_alloc " ++ PPArg, Config, opt,
			  10*60) of
	ok ->
	    OkRes;
	Other ->
	    Other
    end.

old_threads(suite) -> [];
old_threads(doc) ->   ["Checks that sl_alloc function okay in a threaded"
		       "environment"];
old_threads(Config) when list(Config) ->
    {OkRes, PPArg} = threads_tests_modifier(),
    case run_pp_test_case("threads", "old_sl_alloc " ++ PPArg, Config, opt,
			  10*60) of
	ok ->
	    OkRes;
	Other ->
	    Other
    end.

threads_debug(suite) -> [];
threads_debug(doc) ->   ["Checks that sl_alloc function okay in a threaded"
			 "environment (debug compiled)"];
threads_debug(Config) when list(Config) ->
    {OkRes, PPArg} = threads_tests_modifier(),
    case run_pp_test_case("threads", "new_sl_alloc " ++ PPArg, Config,
			  debug, 20*60) of
	ok ->
	    OkRes;
	Other ->
	    Other
    end.

old_threads_debug(suite) -> [];
old_threads_debug(doc) ->   ["Checks that sl_alloc function okay in a "
			     "threaded environment (debug compiled)"];
old_threads_debug(Config) when list(Config) ->
    {OkRes, PPArg} = threads_tests_modifier(),
    case run_pp_test_case("threads",  "old_sl_alloc " ++ PPArg, Config,
			  debug, 20*60) of
	ok ->
	    OkRes;
	Other ->
	    Other
    end.

old_mmap_leak(suite) -> [];
old_mmap_leak(doc) ->   ["Checks that mmapps aren't leaked"];
old_mmap_leak(Config) when list(Config) ->
    run_pp_test_case("old.mmap_leak", "old_sl_alloc", Config, opt, 60).

old_mmap_leak_debug(suite) -> [];
old_mmap_leak_debug(doc) ->   ["Checks that mmapps aren't leaked "
			   "(debug compiled)"];
old_mmap_leak_debug(Config) when list(Config) ->
    run_pp_test_case("old.mmap_leak", "old_sl_alloc", Config, debug, 2*60).

old_mmap_max(suite) -> [];
old_mmap_max(doc) ->   ["Checks that mmap_max is enforced"];
old_mmap_max(Config) when list(Config) ->
    run_pp_test_case("old.mmap_max", "old_sl_alloc", Config, opt, 60).

old_mmap_max_debug(suite) -> [];
old_mmap_max_debug(doc) ->   ["Checks that mmap_max is enforced "
			  "(debug compiled)"];
old_mmap_max_debug(Config) when list(Config) ->
    run_pp_test_case("old.mmap_max", "old_sl_alloc", Config, debug, 2*60).

halloc_wrap(suite) -> [];
halloc_wrap(doc) ->   ["Checks that the HAlloc() makro doesn't fail due "
		       "to pointer arithmetics wrapping around"];
halloc_wrap(Config) when list(Config) ->
    run_pp_test_case("halloc_wrap", Config, opt, 60).

%%% Utilities

find_erl_top_in_dir(Dir) ->
    case catch file:list_dir(Dir) of
	{ok, FileNames} ->
	    case catch
		lists:foldl(fun (File, Acc) ->
				    Cand = filename:join(Dir, File),
				    case catch file:list_dir(Cand) of
					{ok, FN} ->
					    case {lists:member("EPLICENCE", FN),
						  lists:member("erts", FN),
						  lists:member("lib", FN),
						  lists:member("bin", FN)} of
						{true, true, true, true} ->
						% Abort if more than one
						% possible ERL_TOP is found.
						    Acc = false,
						    Cand;
						_ ->
						    Acc
					    end;
					_ ->
					    Acc
				    end
			    end,
			    false,
			    FileNames) of
		ERL_TOP when list(ERL_TOP) ->
		    ERL_TOP;
		_ ->
		    false
	    end;
	_ ->
	    false
    end.

find_erl_top() ->
    case os:getenv("ERL_TOP") of
	false ->
	    % Try to find ERL_TOP in ../../
	    % (open source night builds are usally placed here)
	    ERL_TOP_PARENT = filename:join(
			       filename:dirname(code:which(test_server)),
			       filename:join("..", "..")),
	    case find_erl_top_in_dir(ERL_TOP_PARENT) of
		false ->
		    false;
		ERL_TOP when list(ERL_TOP) ->
		    filename:join(ERL_TOP_PARENT, ERL_TOP)
	    end;
	List when list(List) ->
	    List
    end.

print_pp_data(Data) ->
    test_server:format("~s", [Data]).

get_pp_test_result(Port, TimeOut) ->
    get_pp_test_result(Port, TimeOut, "").

get_pp_test_result(Port, TimeOut, AccData) ->
    receive
	{Port, {data, Data}} ->
	    get_pp_test_result(Port, TimeOut, [AccData, Data]);
	{Port,{exit_status, ES}} ->
	    catch port_close(Port),
	    print_pp_data(AccData),
	    ES;
	TimeOut ->
	    catch port_close(Port),
	    print_pp_data(AccData),
	    ?line test_server:fail(timeout)
    end.

make_pp_test_case(SrcDir, TargetDir, ERL_TOP, Type, TestName, TO) ->
    ?line Make =
	find_make() ++ " "
	++ filename:join(TargetDir, TestName)
	++ " ERL_TOP=" ++ ERL_TOP
	++ " TYPE=" ++ Type
	++ " TARGET_DIR=" ++ TargetDir,
    ?line PortSett = [in,
		      use_stdio,
		      stream,
		      stderr_to_stdout,
		      exit_status,
		      {cd, SrcDir}],
    test_server:format("open_port({spawn, ~p}, ~p)~n", [Make, PortSett]),
    case open_port({spawn, Make}, PortSett) of
	Port when port(Port) ->
	    case get_pp_test_result(Port, TO) of
		?PP_SUCCEEDED ->
		    ok;
		?PP_SKIPPED ->
		    {skipped, "Inspect log for reason"};
		ErrorCode ->
		    ?line test_server:fail({"Make failed with error code:",
					    ErrorCode})
	    end;
	Error ->
	    ?line test_server:fail({"Make failed with error:", Error})
    end.

execute_pp_test_case(TestName, PPArg, Dir, TO) ->
    ?line TestProg = filename:join(Dir, TestName) ++ " " ++ PPArg,
    ?line PortSett = [in,
		      use_stdio,
		      stream,
		      stderr_to_stdout,
		      exit_status],
    test_server:format("open_port({spawn, ~p}, ~p)~n",
		       [TestProg, PortSett]),
    case open_port({spawn, TestProg}, PortSett) of
	Port when port(Port) ->
	    case get_pp_test_result(Port, TO) of
		?PP_SUCCEEDED ->
		    ok;
		?PP_SKIPPED ->
		    {skipped, "Inspect log for reason"};
		?PP_FAILED ->
		    ?line test_server:fail("Failed; inspect log for reason");
		Other ->
		    ?line test_server:fail({unknown, Other})
	    end;
	Error ->
	    ?line test_server:fail({error, Error})
    end.


get_pp_make_options(opt) ->
    {make, {"opt", ".opt.test"}};
get_pp_make_options(debug) ->
    {make, {"debug", ".debug.test"}}.

run_pp_test_case(TestName, Config, BuildType, TimeOut) ->
    run_pp_test_case(TestName, "new_sl_alloc", Config, BuildType, TimeOut).

run_pp_test_case(TestName, PPArg, Config, BuildType, TimeOutTime) ->
    DataDir = ?config(data_dir,Config),
    PrivDir = ?config(priv_dir,Config),
    case os:type() of
	{unix, _} ->
	    case get_pp_make_options(BuildType) of
		{make, {Type, Suffix}} ->
		    TestExecutable = TestName ++ Suffix,
		    TO = make_ref(),
		    ?line {ok, TR} = timer:send_after(
				       timer:seconds(TimeOutTime),
				       self(),
				       TO),
		    Res = case find_erl_top() of
			      false ->
				  {skipped, "ERL_TOP not found"};
			      ERL_TOP ->
				  case make_pp_test_case(DataDir,
							 PrivDir,
							 ERL_TOP,
							 Type,
							 TestExecutable,
							 TO) of
				      ok ->
					  execute_pp_test_case(TestExecutable,
							       PPArg,
							       PrivDir,
							       TO);
				      Other ->
					  Other
				  end
			  end,
		    ?line {ok, cancel} = timer:cancel(TR),
		    % Cleanup; in case we just missed it
		    receive TO -> ok after 0 -> ok end,
		    Res;
		{skip, Reason} ->
		    {skipped, Reason}
	    end;
	_ ->
	    {skipped, "Only run on unix"}
    end.

find_make() ->
    find_make(["gmake", "make"]).

find_make([]) ->
    ?line test_server:fail("No make found");
find_make([M|Ms]) ->
    case os:find_executable(M) of
	false ->
	    find_make(Ms);
	Make when list(Make) ->
	    filename:nativename(Make)
    end.

% make_list(N, E) ->
%     make_list(N, E, []).

% make_list(0, E, Acc) ->
%     Acc;
% make_list(N, E, Acc) ->
%     make_list(N-1, E, [E|Acc]).


% start_node(Name) ->
%     start_node(Name, "").
start_node(Name, Args) ->
    ?line Pa = filename:dirname(code:which(?MODULE)),
    ?line test_server:start_node(Name,
				 peer,
				 [{args,
				   Args ++ " " ++ "-pa " ++ Pa},
				  {linked, false}]).

stop_node(Node) ->
    ?line test_server:stop_node(Node).

% hostname() ->
%     from($@, atom_to_list(node())).

% from(H, [H | T]) -> T;
% from(H, [_ | T]) -> from(H, T);
% from(H, []) -> [].


%long_or_short() -> 
%    case net_kernel:longnames() of
%	true -> " -name ";
%	false -> " -sname "
%    end.

fetch(X,L) when list(L) ->
    case lists:keysearch(X,1,L) of
	false ->
	    {not_found,X,L};
	{value, {X,V}} ->
	    V
    end;
fetch(X,L) ->
    {badarg, [X,L]}.



print_sl_alloc_info() ->
    ?t:format("sl_alloc info:~n~p~n", [erlang:system_info(sl_alloc)]).

get_nodenames(N, T) ->
    get_nodenames(N, T, []).

get_nodenames(0, _, Acc) ->
    Acc;
get_nodenames(N, T, Acc) ->
    {A, B, C} = now(),
    get_nodenames(N-1, T, [list_to_atom(atom_to_list(?MODULE)
					++ "-"
					++ atom_to_list(T)
					++ "-"
					++ integer_to_list(A)
					++ "-"
					++ integer_to_list(B)
					++ "-"
					++ integer_to_list(C)) | Acc]).
	     

% to_string(X) when integer(X) ->
%     integer_to_list(X);
% to_string(X) when atom(X) ->
%     atom_to_list(X);
% to_string(X) when list(X) ->
%     X.
