#include "sl_alloc_test.h"

#define SBC_THRESHOLD 8192
#define MAX_MMAP_CARRIERS 15
#define NO_OF_BLOCKS   25
#define NO_OF_ALLOC_OPS_PER_BLOCK 10

Uint alloc_seq[] = {
  SBC_THRESHOLD,     /* mmap */
  SBC_THRESHOLD*4,   /* mmap to new mmap */
  SBC_THRESHOLD/50,  /* mmap to malloc */
  SBC_THRESHOLD,     /* malloc to mmap */
  0
};

block blocks[NO_OF_BLOCKS] = {{NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq},
			      {NULL, 0, 0, alloc_seq}};

#ifdef DEBUG
#define RBLK(P) CHK_GRD_GET_RBLK((P), "test case")
#else
#define RBLK(P) (P)
#endif

#define DEC_MMAPPED(P)                                                      \
do {                                                                        \
  LOCK;                                                                     \
  if((P) && IS_MMAPPED_BLOCK(RBLK(P)))                                      \
    mmapped--;                                                              \
  UNLOCK;                                                                   \
} while (0)

#define INC_MMAPPED(P)                                                      \
do {                                                                        \
  LOCK;                                                                     \
  if((P) && IS_MMAPPED_BLOCK(RBLK(P)))                                      \
    mmapped++;                                                              \
  UNLOCK;                                                                   \
} while (0)

#define CHK_MMAPPED                                                         \
do {                                                                        \
  LOCK;                                                                     \
  if(mmapped_chunks != mmapped) {                                           \
    printf("mmap() leak! Used mmap chunks: %d Mmapped chunks: %d\n",        \
	   mmapped, mmapped_chunks);                                        \
    FAIL_TEST;                                                              \
  }                                                                         \
  if(mmapped_chunks > MAX_MMAP_CARRIERS) {                                  \
    printf("More mmap than allowed! Mmapped: %d Allowed: %d\n",             \
	   mmapped_chunks,                                                  \
	   MAX_MMAP_CARRIERS);                                              \
    FAIL_TEST;                                                              \
  }                                                                         \
  UNLOCK;                                                                   \
} while (0)


START_TEST
{
  int mmapped = 0;
  int i, j;
  ErtsSlAllocInit esai;

  if (test_new_sl_alloc)
      SKIP_TEST("Cannot test new sl_alloc");

  memset((void *) &esai, 255, sizeof(ErtsSlAllocInit));
  esai.esla = 1;
  esai.eosla = 1;
  esai.sbct = SBC_THRESHOLD;
  esai.mmc = MAX_MMAP_CARRIERS;

  erts_sl_alloc_init(&esai);


  CHK_MMAPPED;

  for(i = 0; i < NO_OF_ALLOC_OPS_PER_BLOCK; i++)
    for(j = 0; j < NO_OF_BLOCKS; j++) {
      DEC_MMAPPED(blocks[j].p);
      alloc_op(&blocks[j], j + 1, 0);
      INC_MMAPPED(blocks[j].p);
      CHK_MMAPPED;
      printf("Mmapped chunks = %d\n", mmapped);
    }

  for(j = 0; j < NO_OF_BLOCKS; j++) {
    DEC_MMAPPED(blocks[j].p);
    alloc_op(&blocks[j], j + 1, 1);
    INC_MMAPPED(blocks[j].p);
    CHK_MMAPPED;
    printf("Mmapped chunks = %d\n", mmapped);
  }

  CHK_MMAPPED;

  printf("Mmapped chunks = %d\n", mmapped);

  mmapped = 0;
  CHK_MMAPPED;

  TEST_SUCCEEDED;
}
END_TEST
