-module(trace_local_SUITE).

-export([basic_test/0, return_test/0, on_and_off_test/0, stack_grow_test/0,
	 info_test/0, delete_test/1, not_run/1]).

-export([slave/1, exported/1, exported_wrap/1, loop/4, apply_slave_async/4]).

%%
%% Define to run outside of test server
%%
%%-define(STANDALONE,1).
 
%%
%% Define for debug output
%%
%%-define(debug,1).
 
-ifdef(STANDALONE).
-define(config(A,B),config(A,B)).
-export([config/2]).
-else.
-include("test_server.hrl").
-endif.
 
-ifdef(debug).
-ifdef(STANDALONE).
-define(line, erlang:display({?MODULE,?LINE}), ).
-endif.
-define(dbgformat(A,B),io:format(A,B)).
-else.
-ifdef(STANDALONE).
-define(line, noop, ).
-endif.
-define(dbgformat(A,B),noop).
-endif.
 
-ifdef(STANDALONE).
config(priv_dir,_) ->
    ".".
-else.
%% When run in test server.
-export([all/1,basic/1, return/1, on_and_off/1, stack_grow/1,info/1, delete/1,
	 init_per_testcase/2, fin_per_testcase/2]).
init_per_testcase(Case, Config) ->
    ?line Dog=test_server:timetrap(test_server:seconds(30)),
    [{watchdog, Dog}|Config].
 
fin_per_testcase(Case, Config) ->
    shutdown(),
    Dog=?config(watchdog, Config),
    test_server:timetrap_cancel(Dog),
    ok.
all(doc) ->
    ["Test tracing of local function calls and return traces."];
all(suite) ->
    case test_server:is_native(?MODULE) of
	true -> [not_run];
	false -> [basic, return, on_and_off, stack_grow, info, delete]
    end.

not_run(Config) when list(Config) -> 
    {skipped,"Native code"}.

basic(suite) ->
    [];
basic(doc) ->
    ["Tests basic local call-trace"];
basic(Config) when list(Config) ->
    basic_test().

return(suite) ->
    [];
return(doc) ->
    ["Tests the different types of return trace"];
return(Config) when list(Config) ->
    return_test().
 
on_and_off(suite) ->
    [];
on_and_off(doc) ->
    ["Tests turning trace parameters on and off, "
     "both for trace and trace_pattern"];
on_and_off(Config) when list(Config) ->
    on_and_off_test().
 
stack_grow(suite) ->
    [];
stack_grow(doc) ->
    ["Tests the stack growth during return traces"];
stack_grow(Config) when list(Config) ->
    stack_grow_test().
 
info(suite) ->
    [];
info(doc) ->
    ["Tests the trace_info BIF"];
info(Config) when list(Config) ->
    info_test().
 
delete(suite) ->
    [];
delete(doc) ->
    ["Tests putting trace on deleted modules"];
delete(Config) when list(Config) ->
    delete_test(Config).
 
-endif.

-define(CT(M,F,A),{trace,_,call,{M,F,A}}).
-define(CTT(M,F,A),{trace_ts,_,call,{M,F,A},{_,_,_}}).
-define(RF(M,F,A,V),{trace,_,return_from,{M,F,A},V}).
-define(RT(M,F,A),{trace,_,return_to,{M,F,A}}).
-define(RTT(M,F,A),{trace_ts,_,return_to,{M,F,A},{_,_,_}}).

-define(EXPECT(List), (fun() -> case collect() of 
				    List ->
					ok;
				    Else ->
					exit({unexpected_result, Else, 
					      {?MODULE,?LINE}})
				end
		       end)()).



basic_test() ->
    ?line setup([call]),
    ?line erlang:trace_pattern({?MODULE,'_','_'},[],[local]),
    ?line [1,1,1,1] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([?CT(?MODULE,exported_wrap,[1]), ?CT(?MODULE,exported,[1]),
		   ?CT(?MODULE,local,[1]),?CT(?MODULE,local2,[1]),
		   ?CT(?MODULE,local_tail,[1])]),
    ?line erlang:trace_pattern({?MODULE,'_','_'},[],[]),
    ?line [1,1,1,1] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([?CT(?MODULE,exported_wrap,[1])]),
    ?line [1,1,1,1] = lambda_slave(fun() ->
					   exported_wrap(1)
				   end),
    ?line ?EXPECT([]), 
    ?line erlang:trace_pattern({?MODULE,'_','_'},[],[local]),
    ?line [1,1,1,1] = lambda_slave(fun() ->
					   exported_wrap(1)
				   end),
    ?line ?EXPECT([?CT(?MODULE,_,_), %% The fun
		   ?CT(?MODULE,exported_wrap,[1]), ?CT(?MODULE,exported,[1]),
		   ?CT(?MODULE,local,[1]),?CT(?MODULE,local2,[1]),
		   ?CT(?MODULE,local_tail,[1])]),
    ?line erlang:trace_pattern({?MODULE,'_','_'},false,[local]),
    ?line shutdown(),
    ok.

return_test() ->
    ?line setup([call]),
    ?line erlang:trace_pattern({?MODULE,'_','_'},[{'_',[],[{return_trace}]}],
			 [local]),
    ?line erlang:trace_pattern({erlang,hash,'_'},[{'_',[],[{return_trace}]}],
			 [local]),
    ?line [1,1,1,1] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([?CT(?MODULE,exported_wrap,[1]), 
		   ?CT(?MODULE,exported,[1]),
		   ?CT(?MODULE,local,[1]),
		   ?CT(?MODULE,local2,[1]),
		   ?CT(?MODULE,local_tail,[1]),
		   ?CT(erlang,hash,[1,1]),
		   ?RF(erlang,hash,2,1),
		   ?RF(?MODULE,local_tail,1,[1,1]),
		   ?RF(?MODULE,local2,1,[1,1]),
		   ?RF(?MODULE,local,1,[1,1,1]),
		   ?RF(?MODULE,exported,1,[1,1,1,1]),
		   ?RF(?MODULE,exported_wrap,1,[1,1,1,1])]),
    ?line shutdown(),
    ?line setup([call,return_to]),
    ?line erlang:trace_pattern({?MODULE,'_','_'},[],
			       [local]),
    ?line erlang:trace_pattern({erlang,hash,'_'},[],
			       [local]),
    ?line [1,1,1,1] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([?CT(?MODULE,exported_wrap,[1]), 
		   ?CT(?MODULE,exported,[1]),
		   ?CT(?MODULE,local,[1]),
		   ?CT(?MODULE,local2,[1]),
		   ?CT(?MODULE,local_tail,[1]),
		   ?CT(erlang,hash,[1,1]),
		   ?RT(?MODULE,local_tail,1),
		   ?RT(?MODULE,local,1),
		   ?RT(?MODULE,exported,1),
		   ?RT(?MODULE,slave,1)]),
    ?line shutdown(),
    ?line setup([call,return_to]),
    ?line erlang:trace_pattern({?MODULE,'_','_'},[{'_',[],[{return_trace}]}],
			       [local]),
    ?line erlang:trace_pattern({erlang,hash,'_'},[{'_',[],[{return_trace}]}],
			       [local]),
    ?line [1,1,1,1] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([?CT(?MODULE,exported_wrap,[1]), 
		   ?CT(?MODULE,exported,[1]),
		   ?CT(?MODULE,local,[1]),
		   ?CT(?MODULE,local2,[1]),
		   ?CT(?MODULE,local_tail,[1]),
		   ?CT(erlang,hash,[1,1]),
		   ?RF(erlang,hash,2,1),
		   ?RT(?MODULE,local_tail,1),
		   ?RF(?MODULE,local_tail,1,[1,1]),
		   ?RF(?MODULE,local2,1,[1,1]),
		   ?RT(?MODULE,local,1),
		   ?RF(?MODULE,local,1,[1,1,1]),
		   ?RT(?MODULE,exported,1),
		   ?RF(?MODULE,exported,1,[1,1,1,1]),
		   ?RF(?MODULE,exported_wrap,1,[1,1,1,1]),
		   ?RT(?MODULE,slave,1)]),
    ?line shutdown(),
    ok.

on_and_off_test() ->
    ?line Pid = setup([call]),
    ?line 1 = erlang:trace_pattern({?MODULE,local_tail,1},[],[local]),
    ?line LocalTail = fun() ->
			      local_tail(1)
		      end,
    ?line [1,1] = lambda_slave(LocalTail),
    ?line ?EXPECT([?CT(?MODULE,local_tail,[1])]),
    ?line erlang:trace(Pid,true,[return_to]),
    ?line [1,1] = lambda_slave(LocalTail),
    ?line ?EXPECT([?CT(?MODULE,local_tail,[1]),
		   ?RT(?MODULE,_,_)]),
    ?line 0 = erlang:trace_pattern({?MODULE,local_tail,1},[],[global]),
    ?line [1,1] = lambda_slave(LocalTail),
    ?line ?EXPECT([]),
    ?line 1 = erlang:trace_pattern({?MODULE,exported_wrap,1},[],[global]),
    ?line [1,1,1,1] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([?CT(?MODULE,exported_wrap,[1])]),
    ?line 1 = erlang:trace_pattern({?MODULE,exported_wrap,1},[],[local]),
    ?line [1,1,1,1] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([?CT(?MODULE,exported_wrap,[1]),
		   ?RT(?MODULE,slave,1)]),
    ?line 1 = erlang:trace_pattern({erlang,hash,2},[],[local]),
    ?line [1,1,1,1] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([?CT(?MODULE,exported_wrap,[1]),
		   ?CT(erlang,hash,[1,1]),
		   ?RT(?MODULE,local_tail,1),
		   ?RT(?MODULE,slave,1)]),
    ?line erlang:trace(Pid,true,[timestamp]),
    ?line [1,1,1,1] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([?CTT(?MODULE,exported_wrap,[1]),
		   ?CTT(erlang,hash,[1,1]),
		   ?RTT(?MODULE,local_tail,1),
		   ?RTT(?MODULE,slave,1)]),
    ?line erlang:trace(Pid,false,[return_to,timestamp]),
    ?line [1,1,1,1] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([?CT(?MODULE,exported_wrap,[1]),
		   ?CT(erlang,hash,[1,1])]),
    ?line erlang:trace(Pid,true,[return_to]),
    ?line 1 = erlang:trace_pattern({erlang,hash,2},[],[]),
    ?line [1,1,1,1] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([?CT(?MODULE,exported_wrap,[1]),
		   ?CT(erlang,hash,[1,1]),
		   ?RT(?MODULE,slave,1)]),
    ?line 1 = erlang:trace_pattern({?MODULE,exported_wrap,1},[],[]),
    ?line [1,1,1,1] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([?CT(?MODULE,exported_wrap,[1]),
		   ?CT(erlang,hash,[1,1])]),
    ?line shutdown(),
    ?line erlang:trace_pattern({'_','_','_'},false,[local]),
    ?line N = erlang:trace_pattern({erlang,'_','_'},true,[local]),
    ?line case erlang:trace_pattern({erlang,'_','_'},false,[local]) of
	      N -> 
		  ok;
	      Else ->
		  exit({number_missmatch, {expected, N}, {got, Else}})
	  end,
    ?line case erlang:trace_pattern({erlang,'_','_'},false,[local]) of
	      N -> 
		  ok;
	      Else2 ->
		  exit({number_missmatch, {expected, N}, {got, Else2}})
	  end,
    ?line M = erlang:trace_pattern({erlang,'_','_'},true,[]),
    ?line case erlang:trace_pattern({erlang,'_','_'},false,[]) of
	      M -> 
		  ok;
	      Else3 ->
		  exit({number_missmatch, {expected, N}, {got, Else3}})
	  end,
    ?line case erlang:trace_pattern({erlang,'_','_'},false,[]) of
	      M -> 
		  ok;
	      Else4 ->
		  exit({number_missmatch, {expected, N}, {got, Else4}})
	  end,
    ok.
    
    
stack_grow_test() ->    
    ?line setup([call,return_to]),
    ?line 1 = erlang:trace_pattern({?MODULE,loop,4},
				   [{'_',[],[{return_trace}]}],[local]),
    ?line Num = 1 bsl 16,
    ?line Fun = 
	fun(F,0) -> 
		ok; 
	   (F,N) -> 
		receive A -> 
			receive B -> 
				receive C ->
					F(F,N-1) 
				end
			end 
		end 
	end, 
    ?line apply_slave_async(?MODULE,loop,[{hej,hopp},[a,b,c],4.5,Num]),
    ?line Fun(Fun,Num + 1),
    ?line ?EXPECT([]), 
    ok.


info_test() ->
    ?line Flags1 = lists:sort([call,return_to]),
    ?line Pid = setup(Flags1),
    ?line Prog = [{['$1'],[{is_integer,'$1'}],[{message, false}]},
		  {'_',[],[]}],
    ?line erlang:trace_pattern({?MODULE,exported_wrap,1},Prog,[local]),
    ?line Self = self(),
    ?line {flags,L} = erlang:trace_info(Pid,flags),
    ?line case lists:sort(L) of
	      Flags1 ->
		  ok;
	      Wrong1 ->
		  exit({bad_result, {erlang,trace_info,[Pid,flags]},
			{expected, Flags1}, {got, Wrong1}})
	  end,
    ?line {tracer,Tracer} = erlang:trace_info(Pid,tracer),
    ?line case Tracer of
	      Self ->
		  ok;
	      Wrong2 ->
		  exit({bad_result, {erlang,trace_info,[Pid,tracer]},
			{expected, Self}, {got, Wrong2}})
	  end,
    ?line {traced,local} = erlang:trace_info({?MODULE,exported_wrap,1},traced),
    ?line {match_spec, MS} = 
	erlang:trace_info({?MODULE,exported_wrap,1},match_spec),
    ?line case MS of
	      Prog ->
		  ok;
	      Wrong3 ->
		  exit({bad_result, {erlang,trace_info,
				     [{?MODULE,exported_wrap,1},
				      match_spec]},
			{expected, Prog}, {got, Wrong3}})
	  end,
    ?line erlang:garbage_collect(self()),
    ?line receive
	  after 1 ->
		  ok
	  end,
    ?line io:format("~p~n",[MS]),
    ?line {match_spec,MS2} = 
	erlang:trace_info({?MODULE,exported_wrap,1},match_spec),
    ?line io:format("~p~n",[MS2]),
    ?line erlang:trace_pattern({?MODULE,exported_wrap,1},[],[]),
    ?line {traced,global} = 
	erlang:trace_info({?MODULE,exported_wrap,1},traced),
    ?line {match_spec,[]} = 
	erlang:trace_info({?MODULE,exported_wrap,1},match_spec),
    ?line {traced,undefined} = 
	erlang:trace_info({?MODULE,exported_wrap,2},traced),
    ?line {match_spec,undefined} = 
	erlang:trace_info({?MODULE,exported_wrap,2},match_spec),
    ?line {traced,false} = erlang:trace_info({?MODULE,exported,1},traced),
    ?line {match_spec,false} = 
	erlang:trace_info({?MODULE,exported,1},match_spec),
    ?line shutdown(),
    ok.

delete_test(Config) ->
    ?line Priv = ?config(priv_dir, Config),
    ?line Data = ?config(data_dir, Config),
    ?line File = filename:join(Data, "trace_local_dummy"),
    ?line Code = filename:join(Priv, "trace_local_dummy"),
    ?line {ok,trace_local_dummy} = c:c(File, [{outdir,Priv}]),
    ?line code:purge(trace_local_dummy),
    ?line code:delete(trace_local_dummy),
    ?line 0 = erlang:trace_pattern({trace_local_dummy,'_','_'},true,[local]),
    ok.


loop(D1,D2,D3,0) ->
    io:format("~p~n",[[D1,D2,D3]]),
    0;
loop(D1,D2,D3,N) ->
    max(N,loop(D1,D2,D3,N-1)).

max(A,B) when A > B ->
    A;
max(A,B) ->
    B.

exported_wrap(Val) ->
    exported(Val).

exported(Val) ->
    [Val | local(Val)]. %% Non tail recursive local call

local(Val) ->
    [Val | local2(Val)]. %% Non tail recursive local call

local2(Val) ->
    local_tail(Val). %% Tail recursive call

local_tail(Val) ->
    [Val , erlang:hash(1,1)]. 





slave(Sync) ->
    Sync ! sync,
    receive
	{From,apply, M, F, A} ->
	    ?line ?dbgformat("Apply: ~p:~p/~p (~p)~n",[M,F,length(A),A]),
	    ?line Res = apply(M,F,A),
	    ?line ?dbgformat("done Apply: ~p:~p/~p (~p)~n",[M,F,length(A),A]),
	    From ! {apply, Res},
	    erlang:trace_pattern({?MODULE,slave,'_'},false,[local]),
	    slave(From);
	{From, lambda, Fun} ->
	    Res = Fun(),
	    From ! {lambda, Res},
	    erlang:trace_pattern({?MODULE,slave,'_'},false,[local]),
	    slave(From);
	die ->
	    exit(normal)
    end.

collect() ->
    L = lists:reverse(collect([])),
    ?dbgformat("Got: ~p~n",[L]),
    L.

collect(A) ->
    receive
	Mess ->
	    collect([Mess | A])
    after 0 ->
	    A
    end.

setup(Flags) ->
    erlang:trace_pattern({'_','_','_'},false,[]),
    erlang:trace_pattern({'_','_','_'},false,[local]),
    Pid = spawn(?MODULE,slave,[self()]),
    receive sync -> ok end,
    put(slave,Pid),
    erlang:trace(Pid, true, Flags),
    Pid.

shutdown() ->
    Pid = get(slave),
    case (catch is_process_alive(Pid)) of
	true ->
	    Ref = erlang:monitor(process,Pid),
	    Pid ! die,
	    receive
		{'DOWN',Ref,process,Pid,_} ->
		    ok
	    end;
	_ ->
	    ok
    end.

apply_slave_async(M,F,A) ->
    Pid = get(slave),
    spawn(?MODULE,apply_slave_async,[M,F,A,Pid]),
    Pid.

apply_slave_async(M,F,A,Pid) ->
    Pid ! {self(),apply, M, F, A},
    receive
	{apply, Res} ->
	    receive
		sync ->
		    Res
	    end
    end.

apply_slave(M,F,A) ->
    Pid = get(slave),
    Pid ! {self(),apply, M, F, A},
    receive
	{apply, Res} ->
	    receive
		sync ->
		    Res
	    end
    end.

lambda_slave(Fun) ->
    Pid = get(slave),
    Pid ! {self(), lambda, Fun},
    receive
	{lambda, Res} ->
	    receive
		sync ->
		    Res
	    end
    end.
